/*
 * dkfobj.c --
 *
 *	This file implements classes and objects in Tcl8.0a{1,2} which
 *	provides 'variables' associated with a group of commands
 *	together with command rewriting to allow the creation of
 *	operations on those commands.  It does not provide any
 *	operations which directly change the semantics of Tcl, unless
 *	the user tries to use any core Tcl commands that are
 *	implemented by direct compilation.
 *
 * Copyright (c) 1997 Donal Kelvin Fellows
 * All rights reserved.
 *
 * Permission is hereby granted, without written agreement and without
 * license or royalty fees, to use, copy, modify, and distribute this
 * software and its documentation for any purpose, provided that the
 * above copyright notice and the following two paragraphs appear in
 * all copies of this software.
 * 
 * IN NO EVENT SHALL THE AUTHOR BE LIABLE TO ANY PARTY FOR DIRECT,
 * INDIRECT, SPECIAL, INCIDENTAL, OR CONSEQUENTIAL DAMAGES ARISING OUT
 * OF THE USE OF THIS SOFTWARE AND ITS DOCUMENTATION, EVEN IF THE
 * AUTHOR HAS BEEN ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 *
 * THE AUTHOR SPECIFICALLY DISCLAIMS ANY WARRANTIES, INCLUDING, BUT
 * NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND
 * FITNESS FOR A PARTICULAR PURPOSE.  THE SOFTWARE PROVIDED HEREUNDER
 * IS ON AN "AS IS" BASIS, AND THE AUTHOR HAS NO OBLIGATION TO PROVIDE
 * MAINTENANCE, SUPPORT, UPDATES, ENHANCEMENTS, OR MODIFICATIONS.
 *
 * It is noted that the author would like to be informed should this
 * software be used in any commercial system - this is not a
 * requirement and can be ignored, but it would be nice (and
 * encourages the development of enhancements to this and other items
 * of software in the future).  At the time of writing, the author can
 * be reached by electronic mail as <fellowsd@cs.man.ac.uk> or
 * <donal@ugglan.demon.co.uk> and it is expected that these are likely
 * to be the most convenient methods of reaching the author.  It is
 * requested that large messages (including patches and attachments)
 * be directed to the former address, where the turnaround time also
 * tends to be smaller.
 *
 * TODO: Document this file properly.  The amount of code is
 *       non-trivial, and there are a number of places where writing
 *       down a description of the cunning bits would be helpful...
 *
 * $Log: dkfobj.c,v $
 * Revision 1.5  1998/04/08 22:17:31  donal
 * Fixed an utterly evil problem with setting traces/links on
 * properties that were only defined as default values in the class,
 * and not actually present in the object at the time.  The net result
 * of the problem (which originally related to a change in the default
 * reference count for a value returned by Tcl_NewObj() during the
 * Tcl8.0 alphas/betas) was that the value in the class definition was
 * freed and often tended to get a completely unrelated value.  This
 * was the cause of at least one bizarre bug...
 *
 * Revision 1.4  1998/02/22 18:42:22  donal
 * One nasty bug cleared up.  It used to be the case that deleting a
 * command referred to by a method declaration that had been then been
 * called would trigger a bad memory access.  This was due to not
 * setting the contents of a crucial hash table.  *sigh*
 *
 * Revision 1.3  1997/05/18 20:52:05  donal
 * A fair number of bugs fixed after an intense session with purify.  Use
 * also indicated that it was necessary to add a linkvar facility.  This
 * uses code unashamedly cobbled together from the Tcl8.0a2 source
 * distribution.  However, this code is located in a separate file
 * (link.c in this directory) so that the authorship and copyright of
 * this file remains unambiguous.
 *
 * The transfer of the system to Tcl8.0a2 has a few advantages in that it
 * allows the removal of a lot of clutter from the system due to the
 * cleaning up of Tcl's object interfaces.
 *
 * Revision 1.2  1997/03/09 17:23:37  donal
 * Well, we're getting a little closer to the Tcl coding conventions.
 * The ones that are left are either ones I disagree with (especially
 * since I get better support from my editor if I do things differently)
 * or are ones that I haven't got around to complying with yet.
 * Documentation is the prime problem area here, especially when I've got
 * a few functions with very closely related functions which would mean
 * that implementing them as a group would make far more sense.  There
 * are still a few places where the behaviour of the system could be
 * 'interesting' due to the use of method procedure name caching, and yet
 * without the caching the performance stinks (measured with previous
 * versions).  The areas to watch out for are renaming of commands (this
 * doesn't have the conventionally-expected results) and implementing
 * methods with system commands, where there is a nastly little bug in
 * the Tcl system (at least in Tcl8.0a1)
 *
 * Revision 1.1  1997/02/06 21:14:17  donal
 * Initial version.  Much still to do...
 * */

static char rcsid[] = "$Id: dkfobj.c,v 1.5 1998/04/08 22:17:31 donal Exp fellows
d $";

/* Outline of calling conventions from Tcl */

/* The class command
 * -----------------
 * class define              $name      $definition ?$definition ...?
 * class exists              $name
 * class info   class        $class
 * class info   method       $clobj     $method
 * class info   object       $object
 * class info   property     $clobj     $property
 * class list   classes      ?$pattern?
 * class list   methods      $clobj     ?pattern?
 * class list   objects      $class     ?pattern?
 * class list   properties   $clobj     ?pattern?
 * class of                  $object
 */

/* Class definition parameters
 * ---------------------------
 * inherit	$class
 * virtual
 * constructor:
 * object:
 * class:
 * shared:
 * property	$name  $val
 * method	$name  $proc
 */

/* Property access commands
 * ------------------------
 * $clobj -> $prop
 * $clobj -> $prop =         $val
 * $clobj -> $prop append    ?$val ...?
 * $clobj -> $prop increment ?$val ...?
 * $clobj -> $prop lappend   ?$val ...?
 * $clobj -> $prop exists
 * $clobj -> $prop trace     read  ?$script?
 * $clobj -> $prop trace     write ?$script?
 * $clobj -> $prop unset
 * $clobj -> $prop link      ?$varname?
 */

#include <tcl.h>
#include <string.h>
#include <stdarg.h>
#include "link.h"

#if (TCL_RELEASE_LEVEL == 0)
#error You need a more recent version of Tcl (8.0b1+)
#endif

struct CommandInfo;
struct Property;
struct Method;
struct Object;
struct Class;

struct CommandInfo {
  Tcl_ObjCmdProc *objProc;	/* The cached values which we looked
				 * up. */
  ClientData clientData;
  Tcl_CmdDeleteProc *delProc;	/* The deleter which we replaced */
  ClientData delData;
  Tcl_HashTable references;	/* All the Tcl_Objs that reference
				 * this command */
};

enum PropKind {
  PROPERTY_CLASS, PROPERTY_OBJECT, PROPERTY_SHARED
};
struct Property {
  enum PropKind kind;		/* What is this property? */
  Tcl_Obj *name;		/* Name of the property */
  Tcl_Obj *contents;		/* Contents of the property */
  int tracing;			/* Are we performing a trace on this
				 * property now? */
  Tcl_Obj *trace_write_script;	/* Trace script for writes */
  Tcl_Obj *trace_read_script;	/* Trace script for reads */
  Tcl_Obj *writer;		/* What wrote this property last */
  Tcl_Obj *linked_var;		/* What variable have we linked to? */
  int is_linked;		/* Have we linked? */
  Tcl_Interp *interp;		/* Needed for unlinking on property deletion */
};

enum MethKind {
  METHOD_CONSTRUCTOR, METHOD_CLASS, METHOD_OBJECT, METHOD_SHARED
};
struct Method {
  enum MethKind kind;		/* What is this method? */
  Tcl_Obj *name;		/* Name of the method */
  Tcl_Obj *commandName;		/* Command name (generally procedure
				 * name) for use in rewriting engine
				 * when executing this method.
				 * Expected to cache info the command
				 * for faster execution */
};

struct Object {
  Tcl_Obj *self;		/* "Name" of the object (usually
				 * reference to it too) */
  struct Class *myClass;	/* Reference to the object's class */
  Tcl_HashTable properties;	/* The properties of the object */
  Tcl_Obj *built_with_constr;	/* Which constructor built the object? */
  Tcl_Command token;		/* The actual command that the object
				 * is implemented as */
  Tcl_HashTable references;	/* All Tcl_Objs that refer to this
				 * object */
};

struct Class {
  Tcl_Obj *self;		/* Name of the class */
  int virtual;			/* Is the creation of objects from
				 * this class illegal? */
  Tcl_HashTable *classes;	/* All classes */
  Tcl_HashTable methods;	/* The methods of the class */
  Tcl_HashTable properties;	/* The properties of the class */
  Tcl_HashTable objects;	/* The objects of the class */
  Tcl_Obj *parents;		/* What classes did this class inherit
				 * from? */
  Tcl_Interp *creatorInterp;	/* In what interpreter was this class
				 * created (for when deleting objects */
  Tcl_Command token;		/* The actual command that the class
				 * was implemented as */
};

/* ---------------------------------------------------------------------- */

static int MakeCommandName  _ANSI_ARGS_((Tcl_Interp *, Tcl_Obj *));
static int MakeObjectRef    _ANSI_ARGS_((Tcl_Interp *, Tcl_Obj *));
static int MakeDerefMarker  _ANSI_ARGS_((Tcl_Interp *, Tcl_Obj *));
static void EmptyUpdate     _ANSI_ARGS_((Tcl_Obj *));
static void DupCommandName  _ANSI_ARGS_((Tcl_Obj *, Tcl_Obj *));
static void DupObjectRef    _ANSI_ARGS_((Tcl_Obj *, Tcl_Obj *));
static void DupDerefMarker  _ANSI_ARGS_((Tcl_Obj *, Tcl_Obj *));
static void FreeObjectRef   _ANSI_ARGS_((Tcl_Obj *));
static void FreeCommandName _ANSI_ARGS_((Tcl_Obj *));
static void DeleteCommandAndInvalidateReferences _ANSI_ARGS_((ClientData));

static Tcl_ObjType commandNameType = {
  "DKF object system: command name",
  FreeCommandName, DupCommandName, EmptyUpdate, MakeCommandName
};

static Tcl_ObjType objectType = {
  "DKF object system: object",
  FreeObjectRef, DupObjectRef, EmptyUpdate, MakeObjectRef
};

static Tcl_ObjType dereferenceMarkerType = {
  "DKF object system: dereference marker",
  NULL, DupDerefMarker, EmptyUpdate, MakeDerefMarker
};

static int GetObject _ANSI_ARGS_((Tcl_Interp *, Tcl_Obj *,
				  Tcl_HashTable *, struct Object **));
static int DereferenceP _ANSI_ARGS_((Tcl_Interp *interp, Tcl_Obj *obj));

/* ---------------------------------------------------------------------- */

static int ClassAddDefinition _ANSI_ARGS_((Tcl_Interp *interp,
					   struct Class *newClass,
					   Tcl_Obj *defList,
					   Tcl_HashTable *classes));
static int MethodCall _ANSI_ARGS_((Tcl_Interp *interp,
				   Tcl_Obj *theMethod,
				   Tcl_Obj *theOwner,
				   int objc, Tcl_Obj *CONST objv[]));

static int ClassImplCmd _ANSI_ARGS_((ClientData clientData,
				     Tcl_Interp *interp,
				     int objc, Tcl_Obj *CONST objv[]));
static int ObjImplCmd   _ANSI_ARGS_((ClientData clientData,
				     Tcl_Interp *interp,
				     int objc, Tcl_Obj *CONST objv[]));
static int ClassCommand _ANSI_ARGS_((ClientData clientData,
				     Tcl_Interp *interp,
				     int objc, Tcl_Obj *CONST objv[]));

static void DelPropForReal   _ANSI_ARGS_((char *clientData));
static void DelMethodForReal _ANSI_ARGS_((char *clientData));
static void DelClassForReal  _ANSI_ARGS_((char *clientData));
static void DelObjForReal    _ANSI_ARGS_((char *clientData));

static void DelClass    _ANSI_ARGS_((ClientData clientData));
static void DelClassCmd _ANSI_ARGS_((ClientData clientData));
static void DelObjCmd   _ANSI_ARGS_((ClientData clientData));

static int ClassWriteTrace  _ANSI_ARGS_((struct Property *propDefn,
					 Tcl_Interp *interp));
static int ClassReadTrace   _ANSI_ARGS_((struct Property *propDefn,
					 Tcl_Interp *interp));
static int ObjectWriteTrace _ANSI_ARGS_((struct Property *propDefn,
					 Tcl_Interp *interp));
static int ObjectReadTrace  _ANSI_ARGS_((struct Property *propDefn,
					 Tcl_Interp *interp));

static struct Object   *BuildObject  _ANSI_ARGS_((Tcl_Interp *interp,
						  struct Class *clsDefn,
						  Tcl_Obj *creatorConstr));
static struct Property *MakeProperty _ANSI_ARGS_((enum PropKind kind,
						  Tcl_Obj *name,
						  Tcl_Obj *contents,
						  Tcl_Obj *writer,
						  Tcl_Interp *interp));
static int WriteObjectProp _ANSI_ARGS_((Tcl_Interp *interp,
					struct Object *objDefn,
					Tcl_Obj *propName, Tcl_Obj *contents,
					int defaulted, Tcl_HashEntry *hent,
					Tcl_Obj **afterContents));
static int ReadObjectProp  _ANSI_ARGS_((Tcl_Interp *interp,
					struct Object *objDefn,
					Tcl_Obj *propName,
					Tcl_Obj **contentsPtr, int *defaulted,
					Tcl_HashEntry **hentPtr));
static Tcl_HashEntry *FindHashEntry _ANSI_ARGS_((Tcl_HashTable *tablePtr,
						 Tcl_Obj *name));
static Tcl_HashEntry *CreateHashEntry _ANSI_ARGS_((Tcl_HashTable *tablePtr,
						   Tcl_Obj *name));
extern int Dkfobj_Init _ANSI_ARGS_((Tcl_Interp *interp));

/* ---------------------------------------------------------------------- */

static char *propAccesses[] = {
  "=", "append", "exists", "increment", "lappend", "trace", "unset", "link",
  NULL
};
static char *traceCmds[] = {
  "read", "write", NULL
};
static char *defnkeywords[] = {
  "class:", "constructor:", "inherit", "method", "object:", "property",
  "shared:", "virtual", NULL
};
static char *classcmds[] = {
  "define", "exists", "info", "list", "of", NULL
};
static char *classinfos[] = {
  "class", "method", "object", "property", NULL
};
static char *classlists[] = {
  "classes", "methods", "objects", "properties", NULL
};

/* ---------------------------------------------------------------------- */

static void
EmptyUpdate(obj)
     Tcl_Obj *obj;
{
  /* Do nothing at all!  This is a place-holder function... */
}

#ifdef __GNUC__
static __inline__ char *S(Tcl_Obj *obj);
static __inline__ char *
#else
static char *S _ANSI_ARGS_((Tcl_Obj *obj));
static char *
#endif
S(obj)
     Tcl_Obj *obj;
{
  if (obj)
    return Tcl_GetStringFromObj(obj,NULL);
  else
    return NULL;
}

#ifdef __GNUC__
static __inline__ Tcl_Obj *
#else
static Tcl_Obj *
#endif
NewObj(void) {
  Tcl_Obj *o = Tcl_NewObj();
  Tcl_IncrRefCount(o);
  return o;
}


/* ---------------------------------------------------------------------- */

/* Hashing of Tcl_Objs instead of strings
 *
 * Two library routines that should have been in the Tcl library
 * itself.  Indeed, if they were located there, then they could be
 * made to work correctly with embedded NULs...
 *
 * The names are modelled directly after the Tcl operations they sit
 * on top of.
 */
static Tcl_HashEntry *
FindHashEntry(tablePtr, name)
     Tcl_HashTable *tablePtr;
     Tcl_Obj *name;
{
  return Tcl_FindHashEntry(tablePtr, Tcl_GetStringFromObj(name, NULL));
}

static Tcl_HashEntry *
CreateHashEntry(tablePtr, name)
     Tcl_HashTable *tablePtr;
     Tcl_Obj *name;
{
  int isNew;
  return Tcl_CreateHashEntry(tablePtr, Tcl_GetStringFromObj(name, NULL),
			     &isNew);
}

/*
 * ----------------------------------------------------------------------
 *
 * dereferenceMarkerType -- implementation
 *
 * 	Caches the dereference symbol.  Implemented this way for speed.
 *
 * ----------------------------------------------------------------------
 */

static void
DupDerefMarker(src, dup)
     Tcl_Obj *src, *dup;
{
  dup->typePtr = &dereferenceMarkerType;
}
static int
MakeDerefMarker(interp, obj)
     Tcl_Interp *interp;
     Tcl_Obj *obj;
{
  obj->typePtr = &dereferenceMarkerType;
  return TCL_OK;
}
static int
DereferenceP(interp, obj)
     Tcl_Interp *interp;
     Tcl_Obj *obj;
{
  int length;
  char *bytes;
  if (obj->typePtr == &dereferenceMarkerType) return 1;
  bytes = Tcl_GetStringFromObj(obj, &length);
  if (length==2 && bytes[0]=='-' && bytes[1]=='>') {
    Tcl_ConvertToType(interp, obj, &dereferenceMarkerType);
    return 1;
  }
  return 0;
}

/*
 * ----------------------------------------------------------------------
 *
 * objectType -- implementation
 *
 * 	Caches objects for faster lookup of info about the object.
 *
 * ----------------------------------------------------------------------
 */

static void
DupObjectRef(srcPtr, dupPtr)
     Tcl_Obj *srcPtr, *dupPtr;
{
  struct Object *objDefn = (struct Object *)srcPtr->internalRep.otherValuePtr;
  int isNew;

  dupPtr->internalRep.otherValuePtr = srcPtr->internalRep.otherValuePtr;
  dupPtr->typePtr = &objectType;
  Tcl_CreateHashEntry(&objDefn->references, (char *)objDefn, &isNew);
}
static void
FreeObjectRef(objPtr)
     Tcl_Obj *objPtr;
{
  struct Object *objDefn = (struct Object *)objPtr->internalRep.otherValuePtr;
  Tcl_HashEntry *hent = Tcl_FindHashEntry(&objDefn->references,
					  (char *)objDefn);
  if (hent) Tcl_DeleteHashEntry(hent);
}

/* The next line is Tcl-internal, but does what I want! */
EXTERN void panic _ANSI_ARGS_(TCL_VARARGS(char *,format));

static int
MakeObjectRef(interp, objPtr)
     Tcl_Interp *interp;
     Tcl_Obj *objPtr;
{
  panic("MakeObjectRef: some general code tried to use this "
	"(unsupported) interface");
  return TCL_ERROR;
}

static int
GetObject(interp, objPtr, classes, oReturn)
     Tcl_Interp *interp;
     Tcl_Obj *objPtr;
     Tcl_HashTable *classes;
     struct Object **oReturn;
{
  Tcl_HashSearch hsearch;
  Tcl_HashEntry *hentCls, *hentObj;

  if (objPtr->typePtr == &objectType) {
    *oReturn = objPtr->internalRep.otherValuePtr;
    return TCL_OK;
  }
  hentCls = Tcl_FirstHashEntry(classes, &hsearch);
  for (;hentCls; hentCls=Tcl_NextHashEntry(&hsearch)) {
    struct Class *clsDefn = Tcl_GetHashValue(hentCls);
    hentObj = FindHashEntry(&clsDefn->objects, objPtr);
    if (hentObj) {
      *oReturn = Tcl_GetHashValue(hentObj);
      if (objPtr->typePtr && objPtr->typePtr->freeIntRepProc) {
	objPtr->typePtr->freeIntRepProc(objPtr);
      }
      objPtr->typePtr = &objectType;
      objPtr->internalRep.otherValuePtr = *oReturn;
      return TCL_OK;
    }
  }
  Tcl_AppendResult(interp, "unknown object \"", S(objPtr), "\"", NULL);
  return TCL_ERROR;
}

/*
 * ----------------------------------------------------------------------
 *
 * commandNameType -- implementation
 *
 * 	Caches command lookups, correctly dealing with command
 *	deletion.  Doesn't handle command renaming correctly though.
 *
 * ----------------------------------------------------------------------
 */

static void
FreeCommandName(obj)
     Tcl_Obj *obj;
{
  struct CommandInfo *cn = (struct CommandInfo *)
    obj->internalRep.otherValuePtr;
  Tcl_HashEntry *hent = Tcl_FindHashEntry(&cn->references, (char *)obj);
  if (hent) Tcl_DeleteHashEntry(hent);
}

static void
DupCommandName(srcPtr, dupPtr)
     Tcl_Obj *srcPtr, *dupPtr;
{
  struct CommandInfo *cn = (struct CommandInfo *)
    srcPtr->internalRep.otherValuePtr;
  int isNew;

  dupPtr->internalRep.otherValuePtr = srcPtr->internalRep.otherValuePtr;
  dupPtr->typePtr = &commandNameType;
  Tcl_CreateHashEntry(&cn->references, (char *)dupPtr, &isNew);
}

static int
MakeCommandName(interp, obj)
     Tcl_Interp *interp;
     Tcl_Obj *obj;
{
  struct CommandInfo *cn;
  Tcl_CmdInfo info;
  int len;
  char *name = Tcl_GetStringFromObj(obj, NULL);
  Tcl_HashEntry *hent;

  if (!Tcl_GetCommandInfo(interp, name, &info)) {
    Tcl_AppendResult(interp, "invalid command name \"", S(obj), "\"", NULL);
    return TCL_ERROR;
  }

  /* Install our trace on command deletion */
  if (info.deleteProc != DeleteCommandAndInvalidateReferences) {
    cn = (struct CommandInfo *)Tcl_Alloc(sizeof(struct CommandInfo));
    cn->objProc = info.objProc;
    cn->clientData = info.objClientData;
    cn->delProc = info.deleteProc;
    cn->delData = info.deleteData;
    Tcl_InitHashTable(&cn->references, TCL_ONE_WORD_KEYS);
    info.deleteProc = DeleteCommandAndInvalidateReferences;
    info.deleteData = cn;
    Tcl_SetCommandInfo(interp, name, &info);
  } else {
    cn = info.deleteData;
  }

  if (obj->typePtr && obj->typePtr->freeIntRepProc) {
    obj->typePtr->freeIntRepProc(obj);
  }
  obj->internalRep.otherValuePtr = cn;
  hent = Tcl_CreateHashEntry(&cn->references, (char *)obj, &len);
  Tcl_SetHashValue(hent, obj);
  obj->typePtr = &commandNameType;
  return TCL_OK;
}

static void
DeleteCommandAndInvalidateReferences(clientData)
     ClientData clientData;
{
  struct CommandInfo *cn = clientData;
  Tcl_HashSearch hsearch;
  Tcl_HashEntry *hent;

  cn->delProc(cn->delData);
  hent = Tcl_FirstHashEntry(&cn->references, &hsearch);
  for (;hent; hent=Tcl_NextHashEntry(&hsearch)) {
    Tcl_Obj *obj = Tcl_GetHashValue(hent);
    /* Invalidate the reference in the object */
    obj->typePtr = NULL;
  }
  Tcl_DeleteHashTable(&cn->references);
  Tcl_Free((char *)cn);
}

static int
MethodCall(interp, theMethod, theOwner, objc, objv)
     Tcl_Interp *interp;
     Tcl_Obj *theMethod, *theOwner;
     int objc;
     Tcl_Obj *CONST objv[];
{
  struct CommandInfo *cn;
  Tcl_Obj **vector;
  int retcode;

  if (theMethod->typePtr != &commandNameType) {
    if (Tcl_ConvertToType(interp, theMethod, &commandNameType))
      return TCL_ERROR;
  }
  cn = (struct CommandInfo *)theMethod->internalRep.otherValuePtr;
  vector = (Tcl_Obj **)Tcl_Alloc((objc+2)*sizeof(Tcl_Obj *));
  vector[0] = theMethod;
  vector[1] = theOwner;
  memcpy(vector+2, objv, objc*sizeof(Tcl_Obj *));
  retcode = cn->objProc(cn->clientData, interp, objc+2, vector);
  Tcl_Free((char *)vector);
  return retcode;
}

/* ---------------------------------------------------------------------- */

/* Traces on class properties */

static int
ClassWriteTrace(propDefn, interp)
     struct Property *propDefn;
     Tcl_Interp *interp;
{
  if (propDefn->is_linked) {
    Dkfobj_UpdateLinkedVar(interp, propDefn->linked_var);
  }
  if (!propDefn->tracing) {
    int code;
    propDefn->tracing = 1;
    code = Tcl_GlobalEvalObj(interp, propDefn->trace_write_script);
    propDefn->tracing = 0;
    if (code != TCL_OK) {
      Tcl_AddObjErrorInfo(interp, "\n    (write trace on class property)", -1);
      return TCL_ERROR;
    }
  }
  return TCL_OK;
}

static int
ClassReadTrace(propDefn, interp)
     struct Property *propDefn;
     Tcl_Interp *interp;
{
  if (!propDefn->tracing) {
    int code;
    propDefn->tracing = 1;
    code = Tcl_GlobalEvalObj(interp, propDefn->trace_read_script);
    propDefn->tracing = 0;
    if (code != TCL_OK) {
      Tcl_AddObjErrorInfo(interp, "\n    (read trace on class property)", -1);
      return TCL_ERROR;
    }
  }
  return TCL_OK;
}

/* Traces on object properties */

static int
ObjectWriteTrace(propDefn, interp)
     struct Property *propDefn;
     Tcl_Interp *interp;
{
  if (!propDefn->tracing) {
    int code;
    propDefn->tracing = 1;
    code = Tcl_GlobalEvalObj(interp, propDefn->trace_write_script);
    propDefn->tracing = 0;
    if (code != TCL_OK) {
      Tcl_AddObjErrorInfo(interp,"\n    (write trace on object property)", -1);
      return TCL_ERROR;
    }
  }
  return TCL_OK;
}

static int
ObjectReadTrace(propDefn, interp)
     struct Property *propDefn;
     Tcl_Interp *interp;
{
  if (!propDefn->tracing) {
    int code;
    propDefn->tracing = 1;
    code = Tcl_GlobalEvalObj(interp, propDefn->trace_read_script);
    propDefn->tracing = 0;
    if (code != TCL_OK) {
      Tcl_AddObjErrorInfo(interp, "\n    (read trace on object property)", -1);
      return TCL_ERROR;
    }
  }
  return TCL_OK;
}

/* ---------------------------------------------------------------------- */

/* create an object - does not call constructor(s) */

static struct Object *
BuildObject(interp, clsDefn, creatorConstr)
     Tcl_Interp *interp;
     struct Class *clsDefn;
     Tcl_Obj *creatorConstr;
{
  static int uniqueNumber = 0;
  Tcl_Obj *num;
  Tcl_HashEntry *hent;
  int isNew, slen;
  struct Object *objDefn = (struct Object *)Tcl_Alloc(sizeof(struct Object));
  char *bytes = Tcl_GetStringFromObj(clsDefn->self, &slen);

  /* This is where things get interesting.  The idea behind doing all
   * this here is to avoid a costly search which a) would return a
   * value we already know, and b) would actually free because we
   * haven't completed that stage yet!
   *
   * We do all this by taking a string obj (actually untyped) and
   * modifying it.  Simple!
   */
  objDefn->self = Tcl_NewStringObj(bytes, slen);
  num = Tcl_NewIntObj(uniqueNumber++);
  Tcl_AppendToObj(objDefn->self, Tcl_GetStringFromObj(num, NULL), -1);
  Tcl_DecrRefCount(num);
  objDefn->self->internalRep.otherValuePtr = objDefn;
  objDefn->self->typePtr = &objectType;

  objDefn->myClass = clsDefn;
  Tcl_InitHashTable(&objDefn->properties, TCL_STRING_KEYS);
  Tcl_InitHashTable(&objDefn->references, TCL_ONE_WORD_KEYS);
  Tcl_CreateHashEntry(&objDefn->references, (char *)objDefn->self, &isNew);
  objDefn->built_with_constr = creatorConstr;
  objDefn->token = Tcl_CreateObjCommand(interp, objDefn->self->bytes,
					ObjImplCmd, objDefn, DelObjCmd);
  hent = Tcl_CreateHashEntry(&clsDefn->objects, objDefn->self->bytes, &isNew);
  Tcl_SetHashValue(hent, objDefn);
  return objDefn;
}

/* The implementation of a class - see documentation elsewhere */

static int
ClassImplCmd(clientData, interp, objc, objv)
     ClientData clientData;
     Tcl_Interp *interp;
     int objc;
     Tcl_Obj *CONST objv[];
{
  struct Class *clsDefn = clientData;

  if (objc<2) {
    /* Default body-less constructor */
    struct Object *objDefn;

    if (clsDefn->virtual) {
      Tcl_AppendResult(interp, "class \"", S(objv[0]), "\" is virtual", NULL);
      return TCL_ERROR;
    }
    objDefn = BuildObject(interp, clsDefn, NewObj());
    Tcl_IncrRefCount(objDefn->self);
    Tcl_SetObjResult(interp, objDefn->self);
    return TCL_OK;
  } else if (DereferenceP(interp, objv[1])) {
    /* Property access */
    if (objc<3) {
      Tcl_AppendResult(interp, "wrong # args: usage \"", S(objv[0]),
		       " -> propName ?operation ...?\"", NULL);
      return TCL_ERROR;
    } else {
      Tcl_HashEntry *hent = FindHashEntry(&clsDefn->properties, objv[2]);
      struct Property *propDefn;
      if (objc==3) {
	if (!hent) {
	unknownProperty:
	  Tcl_AppendResult(interp, "unknown property \"", S(objv[2]),
			   "\" for class \"", S(objv[0]), "\"", NULL);
	  return TCL_ERROR;
	}
	propDefn = Tcl_GetHashValue(hent);
	/* Read a property */
	Tcl_Preserve(propDefn);
	if (ClassReadTrace(propDefn, interp) == TCL_OK) {
	  Tcl_SetObjResult(interp, propDefn->contents);
	  Tcl_Release(propDefn);
	  return TCL_OK;
	} else {
	  Tcl_Release(propDefn);
	  return TCL_ERROR;
	}
      } else {
	/* Do something elaborate to a property */
	int i,index;
	long accumulator,addValue;
	Tcl_Obj *tmp;

	if (Tcl_GetIndexFromObj(interp, objv[3], propAccesses,
				"property access subcommand",
				0, &index)!=TCL_OK) {
	  return TCL_ERROR;
	}

	/* Pre-run */
	switch (index) {
	case 2: /* exists - Check if a property exists */
	  if (objc!=4) {
	    Tcl_AppendResult(interp, "wrong # args: should be \"", S(objv[0]),
			     " -> ", S(objv[2]), " exists\"", NULL);
	    return TCL_ERROR;
	  }
	  Tcl_SetObjResult(interp, Tcl_NewBooleanObj(hent!=NULL));
	  return TCL_OK;

	case 6: /* unset - Delete a property - illegal in classes */
	  if (objc!=4) {
	    Tcl_AppendResult(interp, "wrong # args: should be \"", S(objv[0]),
			     " -> ", S(objv[2]), " unset\"", NULL);
	  } else {
	    Tcl_AppendResult(interp, "properties cannot be deleted from"
			     " a class", NULL);
	  }
	  return TCL_ERROR;

	default:
	  if (!hent) goto unknownProperty; /* You cannot create new
					    * properties in a class at
					    * all... */
	  propDefn = Tcl_GetHashValue(hent);
	}

	switch (index) {
	case 0: /* = - Write a property */
	  if (objc!=5) {
	    Tcl_AppendResult(interp, "wrong # args: should be \"", S(objv[0]),
			     " -> ", S(objv[2]), " = value\"", NULL);
	    return TCL_ERROR;
	  }
	  if (propDefn->kind == PROPERTY_OBJECT) {
	    Tcl_AppendResult(interp, "property \"", S(objv[2]),
			     "\" is not writable in class \"", S(objv[0]),
			     "\"", NULL);
	    return TCL_ERROR;
	  }
	  tmp = propDefn->contents;
	  Tcl_IncrRefCount(objv[4]);
	  propDefn->contents = objv[4];
	  Tcl_Preserve(propDefn);
	  if (ClassWriteTrace(propDefn, interp) != TCL_OK) {
	    Tcl_DecrRefCount(objv[4]);
	    propDefn->contents = tmp;
	    Tcl_Release(propDefn);
	    return TCL_ERROR;
	  }
	  Tcl_IncrRefCount(objv[0]);
	  propDefn->writer = objv[0];
	  Tcl_DecrRefCount(tmp);
	  Tcl_SetObjResult(interp, propDefn->contents);
	  Tcl_Release(propDefn);
	  return TCL_OK;

	case 1: /* append - Append to a property */
	  if (propDefn->kind == PROPERTY_OBJECT) {
	    Tcl_AppendResult(interp, "property \"", S(objv[2]),
			     "\" is not writable in class \"", S(objv[0]),
			     "\"", NULL);
	    return TCL_ERROR;
	  }
	  Tcl_Preserve(propDefn);
	  if (ClassReadTrace(propDefn, interp) != TCL_OK) {
	    Tcl_Release(propDefn);
	    return TCL_ERROR;
	  }
	  tmp = propDefn->contents;
	  propDefn->contents = Tcl_DuplicateObj(tmp);
	  for (i=4 ; i<objc ; i++) {
	    int length;
	    char *bytes = Tcl_GetStringFromObj(objv[i], &length);

	    Tcl_AppendToObj(propDefn->contents, bytes, length);
	  }
	  if (ClassWriteTrace(propDefn, interp) != TCL_OK) {
	    Tcl_DecrRefCount(propDefn->contents);
	    propDefn->contents = tmp;
	    Tcl_Release(propDefn);
	    return TCL_ERROR;
	  }
	  Tcl_IncrRefCount(objv[0]);
	  propDefn->writer = objv[0];
	  Tcl_DecrRefCount(tmp);
	  Tcl_IncrRefCount(propDefn->contents);
	  Tcl_SetObjResult(interp, propDefn->contents);
	  Tcl_Release(propDefn);
	  return TCL_OK;


	case 3: /* increment - Add to a property */
	  if (propDefn->kind == PROPERTY_OBJECT) {
	    Tcl_AppendResult(interp, "property \"", S(objv[2]),
			     "\" is not writable in class \"", S(objv[0]),
			     "\"", NULL);
	    return TCL_ERROR;
	  }
	  Tcl_Preserve(propDefn);
	  if (ClassReadTrace(propDefn, interp) != TCL_OK) {
	    Tcl_Release(propDefn);
	    return TCL_ERROR;
	  }
	  accumulator = addValue = 0;
	  if (Tcl_GetLongFromObj(interp, propDefn->contents,
				&accumulator) != TCL_OK) {
	    Tcl_Release(propDefn);
	    return TCL_ERROR;
	  }
	  for (i=4 ; i<objc ; i++) {
	    if (Tcl_GetLongFromObj(interp, objv[i], &addValue) != TCL_OK) {
	      Tcl_Release(propDefn);
	      return TCL_ERROR;
	    }
	    accumulator += addValue;
	  }
	  tmp = propDefn->contents;
	  propDefn->contents = Tcl_NewLongObj(accumulator);
	  Tcl_IncrRefCount(propDefn->contents);
	  if (ClassWriteTrace(propDefn, interp) != TCL_OK) {
	    Tcl_DecrRefCount(propDefn->contents);
	    propDefn->contents = tmp;
	    Tcl_Release(propDefn);
	    return TCL_ERROR;
	  }
	  Tcl_IncrRefCount(objv[0]);
	  propDefn->writer = objv[0];
	  Tcl_DecrRefCount(tmp);
	  Tcl_SetObjResult(interp, propDefn->contents);
	  Tcl_Release(propDefn);
	  return TCL_OK;

	case 4: /* lappend - Append list items to a property */
	  if (propDefn->kind == PROPERTY_OBJECT) {
	    Tcl_AppendResult(interp, "property \"", S(objv[2]),
			     "\" is not writable in class \"", S(objv[0]),
			     "\"", NULL);
	    return TCL_ERROR;
	  }
	  Tcl_Preserve(propDefn);
	  if (ClassReadTrace(propDefn, interp) != TCL_OK) {
	    Tcl_Release(propDefn);
	    return TCL_ERROR;
	  }
	  tmp = propDefn->contents;
	  propDefn->contents = Tcl_DuplicateObj(tmp);
	  for (i=4 ; i<objc ; i++) {
	    if (Tcl_ListObjAppendElement(interp, propDefn->contents,
					 objv[i]) != TCL_OK) {
	      Tcl_DecrRefCount(propDefn->contents);
	      propDefn->contents = tmp;
	      Tcl_Release(propDefn);
	      return TCL_ERROR;
	    }
	  }
	  if (ClassWriteTrace(propDefn, interp) != TCL_OK) {
	    Tcl_DecrRefCount(propDefn->contents);
	    propDefn->contents = tmp;
	    Tcl_Release(propDefn);
	    return TCL_ERROR;
	  }
	  Tcl_IncrRefCount(objv[0]);
	  propDefn->writer = objv[0];
	  Tcl_DecrRefCount(tmp);
	  Tcl_IncrRefCount(propDefn->contents);
	  Tcl_SetObjResult(interp, propDefn->contents);
	  Tcl_Release(propDefn);
	  return TCL_OK;

	case 5: /* trace - Set/query a trace a property */
	  if (objc<5 || objc>6) {
	    Tcl_AppendResult(interp, "wrong # args: should be \"", S(objv[0]),
			     " -> ", S(objv[2]), " trace read/write "
			     "?script?\"", NULL);
	    return TCL_ERROR;
	  } else {
	    static char *pts[] = {"read", "write", NULL};
	    if (Tcl_GetIndexFromObj(interp, objv[4], pts,
				    "property trace subcommand",
				    0, &index) != TCL_OK) {
	      return TCL_ERROR;
	    }
	  }
	  switch (index) {
	  case 0:
	    if (objc==6) {
	      Tcl_DecrRefCount(propDefn->trace_read_script);
	      propDefn->trace_read_script = objv[5];
	      Tcl_IncrRefCount(propDefn->trace_read_script);
	    }
	    Tcl_SetObjResult(interp, propDefn->trace_read_script);
	    return TCL_OK;
	  case 1:
	    if (objc==6) {
	      Tcl_DecrRefCount(propDefn->trace_write_script);
	      propDefn->trace_write_script = objv[5];
	      Tcl_IncrRefCount(propDefn->trace_write_script);
	    }
	    Tcl_SetObjResult(interp, propDefn->trace_write_script);
	    return TCL_OK;
	  default:
	    panic("switch fallthrough");
	  }

	case 7: /* link - mirror a property in a global variable */
	  if (objc<4 || objc>5) {
	    Tcl_AppendResult(interp, "wrong # args: should be \"", S(objv[0]),
			     " -> ", S(objv[2]), " link ?variable?\"", NULL);
	  }
	  propDefn = Tcl_GetHashValue(hent);
	  if (objc==5) {
	    int vlen;

	    Tcl_GetStringFromObj(objv[4], &vlen);

	    if (vlen) {
	      if (Dkfobj_LinkVar(interp, objv[4], &propDefn->contents,
				 propDefn->kind == PROPERTY_OBJECT)!=TCL_OK) {
		return TCL_ERROR;
	      }
	    }
	    if (propDefn->is_linked) {
	      Dkfobj_UnlinkVar(interp, propDefn->linked_var);
	      propDefn->is_linked = 0;
	    }
	    Tcl_DecrRefCount(propDefn->linked_var);
	    propDefn->linked_var = objv[4];
	    Tcl_IncrRefCount(propDefn->linked_var);
	    if (vlen) {
	      propDefn->is_linked = 1;
	    }
	  }
	  Tcl_SetObjResult(interp, propDefn->linked_var);
	  return TCL_OK;

	default:
	  Tcl_AppendResult(interp, "switch fall-through", NULL);
	  return TCL_ERROR;
	}
      }
    }
  } else {
    /* Method call */
    Tcl_HashEntry *hent = FindHashEntry(&clsDefn->methods, objv[1]);

    if (!hent) {
      Tcl_AppendResult(interp, "unknown method \"", S(objv[1]),
		       "\" for class \"", S(objv[0]), "\"", NULL);
      return TCL_ERROR;
    } else {
      struct Method *mthdDefn = Tcl_GetHashValue(hent);
      struct Object *objDefn;
      int code;
      switch (mthdDefn->kind) {
      case METHOD_CONSTRUCTOR:
	if (clsDefn->virtual) {
	  Tcl_AppendResult(interp, "class \"", S(objv[0]), "\" is virtual",
			   NULL);
	  return TCL_ERROR;
	}
	Tcl_IncrRefCount(objv[1]);
	objDefn = BuildObject(interp, clsDefn, objv[1]);
	Tcl_Preserve(objDefn);
	if (MethodCall(interp, mthdDefn->commandName, objDefn->self,
		       objc-2, objv+2) != TCL_OK) {
	  Tcl_Obj *result = Tcl_GetObjResult(interp);
	  Tcl_Obj *vName = Tcl_NewStringObj("errorInfo", -1);
	  Tcl_Obj *errorInfo;
	  int len;
	  char *msg;

	  Tcl_IncrRefCount(vName);
	  errorInfo = Tcl_ObjGetVar2(interp, vName, NULL, TCL_GLOBAL_ONLY);
	  msg = Tcl_GetStringFromObj(errorInfo, &len);
	  Tcl_DecrRefCount(vName);
	  Tcl_IncrRefCount(result);
	  Tcl_DeleteCommand(interp,
			    Tcl_GetCommandName(interp, objDefn->token));
	  Tcl_Release(objDefn);
	  Tcl_SetObjResult(interp, NULL);
	  Tcl_AddObjErrorInfo(interp, msg, len);
	  Tcl_AddObjErrorInfo(interp, "\n    (during constructor call)", -1);
	  Tcl_DecrRefCount(errorInfo);
	  Tcl_SetObjResult(interp, result);
	  return TCL_ERROR;
	}
	Tcl_IncrRefCount(objDefn->self);
	Tcl_SetObjResult(interp, objDefn->self);
	Tcl_Release(objDefn);
	return TCL_OK;
      case METHOD_CLASS:
      case METHOD_SHARED:
	code = MethodCall(interp, mthdDefn->commandName, clsDefn->self,
			  objc-2, objv+2);
	if (code != TCL_OK) {
	  Tcl_AddObjErrorInfo(interp, "\n    (during class method call)", -1);
	}
	return code;
      case METHOD_OBJECT:
      default: /* Shouldn't happen */
	Tcl_AppendResult(interp, "method \"", S(objv[1]),
			 "\" is not legally callable for class \"",
			 S(clsDefn->self), "\"", NULL);
	return TCL_ERROR;
      }
    }
  }
}

/* ---------------------------------------------------------------------- */

/* Read an object's property, passing back info to make writing the
 * property back efficient */

static int
ReadObjectProp(interp, objDefn, propName, contentsPtr, defaulted, hentPtr)
     Tcl_Interp *interp;
     struct Object *objDefn;
     Tcl_Obj *propName, **contentsPtr;
     int *defaulted;
     Tcl_HashEntry **hentPtr;
{
  Tcl_HashEntry *hent = FindHashEntry(&objDefn->properties, propName);
  if (hent) {
    struct Property *propDefn = Tcl_GetHashValue(hent);
    Tcl_Preserve(propDefn);
    if (ObjectReadTrace(propDefn, interp) != TCL_OK) {
      Tcl_Release(propDefn);
      return TCL_ERROR;
    }
    *hentPtr = hent;
    Tcl_IncrRefCount(propDefn->contents);
    *contentsPtr = propDefn->contents;
    *defaulted = 0;
    Tcl_Release(propDefn);
    return TCL_OK;
  }
  hent = FindHashEntry(&objDefn->myClass->properties, propName);
  if (hent) {
    struct Property *propDefn = Tcl_GetHashValue(hent);
    switch (propDefn->kind) {
    case PROPERTY_SHARED:
    case PROPERTY_OBJECT:
      Tcl_Preserve(propDefn);
      if (ClassReadTrace(propDefn, interp) != TCL_OK) {
	Tcl_Release(propDefn);
	return TCL_ERROR;
      }
      *hentPtr = hent;
      Tcl_IncrRefCount(propDefn->contents);
      *contentsPtr = propDefn->contents;
      *defaulted = (propDefn->kind == PROPERTY_OBJECT);
      Tcl_Release(propDefn);
      return TCL_OK;
    case PROPERTY_CLASS:
    default:
      Tcl_AppendResult(interp, "property \"", S(propName),
		       "\" is unreadable in object \"", S(objDefn->self),
		       "\"", NULL);
      return TCL_ERROR;
    }
  }
  Tcl_AppendResult(interp, "unknown property \"", S(propName),
		   "\" for object \"", S(objDefn->self), "\"", NULL);
  return TCL_ERROR;
}

static struct Property *
MakeProperty(kind, name, contents, writer, interp)
     enum PropKind kind;
     Tcl_Obj *name, *contents, *writer;
     Tcl_Interp *interp;
{
  struct Property *prop = (struct Property *)
    Tcl_Alloc(sizeof(struct Property));

  prop->kind = kind;
  Tcl_IncrRefCount(prop->name     = name);
  Tcl_IncrRefCount(prop->contents = contents);
  Tcl_IncrRefCount(prop->writer   = writer);
  prop->interp = interp;

  prop->tracing = 0;
  prop->is_linked = 0;
  prop->trace_write_script = NewObj();
  prop->trace_read_script = NewObj();
  prop->linked_var = NewObj();

  return prop;
}

/* Write an object's property.  Uses info from ReadObjectProp() and
 * returns a reference to the new property contents (which might not
 * be what is expected due to traces) with the reference count
 * incremented */

static int
WriteObjectProp(interp, objDefn, propName, contents, defaulted, hent,
		afterContents)
     Tcl_Interp *interp;
     struct Object *objDefn;
     Tcl_Obj *propName, *contents, **afterContents;
     int defaulted;
     Tcl_HashEntry *hent;
{
  if (hent) {
    if (defaulted) {
      /* We have a read-only property in the class, but we can create
       * a writable property in the object with the same name.  Do it!
       */
      struct Property *oldP = Tcl_GetHashValue(hent);
      struct Property *propDefn = MakeProperty(PROPERTY_OBJECT,
					       oldP->name, contents,
					       objDefn->self, interp);
      hent = CreateHashEntry(&objDefn->properties, propName);
      Tcl_SetHashValue(hent, propDefn);
      if (afterContents) *afterContents = contents;
      /* There are no traces set, so there's nothing to call here! */
      return TCL_OK;
    } else {
      /* We have a writable property - write it! */
      struct Property *propDefn = Tcl_GetHashValue(hent);
      Tcl_Obj *tmp = propDefn->contents;
      propDefn->contents = contents;
      Tcl_Preserve(objDefn);
      Tcl_Preserve(propDefn);
      if (propDefn->is_linked) {
	Dkfobj_UpdateLinkedVar(interp, propDefn->linked_var);
      }
      if (ObjectWriteTrace(propDefn, interp) != TCL_OK) {
	propDefn->contents = tmp;
	Tcl_Release(propDefn);
	Tcl_Release(objDefn);
	return TCL_ERROR;
      }

      Tcl_DecrRefCount(propDefn->writer);
      Tcl_IncrRefCount(objDefn->self);
      propDefn->writer = objDefn->self;

      Tcl_DecrRefCount(tmp);
      if (afterContents) *afterContents = propDefn->contents;
      Tcl_IncrRefCount(propDefn->contents);
      Tcl_Release(propDefn);
      Tcl_Release(objDefn);
      return TCL_OK;
    }
  } else {
    /* This is an all-new property that we are writing, but we must
     * check that we aren't actually writing over what should be a
     * class-only variable.  This is where things get interesting...
     */
    struct Property *propDefn, *classProp;
    Tcl_Obj *tmp;

    hent = FindHashEntry(&objDefn->myClass->properties, propName);
    if (!hent) {
      /* There is definitely no impediment, so roll with this */
    createAProperty:
      propDefn = MakeProperty(PROPERTY_OBJECT, propName, contents,
			      objDefn->self, interp);
      hent = CreateHashEntry(&objDefn->properties, propName);
      Tcl_SetHashValue(hent, propDefn);
      if (afterContents) *afterContents = contents;
      return TCL_OK;
    }
    /* Hmm.  It exists in the class.  Check what to do */
    classProp = Tcl_GetHashValue(hent);
    switch (classProp->kind) {
    case PROPERTY_SHARED:
      /* Ah, we need to update the class property... */
      tmp = classProp->contents;
      classProp->contents = contents;
      Tcl_Preserve(objDefn);
      Tcl_Preserve(classProp);
      if (classProp->is_linked) {
	Dkfobj_UpdateLinkedVar(interp, classProp->linked_var);
      }
      if (ObjectWriteTrace(classProp, interp) != TCL_OK) {
	classProp->contents = tmp;
	Tcl_Release(classProp);
	Tcl_Release(objDefn);
	return TCL_ERROR;
      }

      Tcl_DecrRefCount(classProp->writer);
      Tcl_IncrRefCount(objDefn->self);
      classProp->writer = objDefn->self;

      Tcl_DecrRefCount(tmp);
      if (afterContents) *afterContents = classProp->contents;
      Tcl_IncrRefCount(classProp->contents);
      Tcl_Release(classProp);
      Tcl_Release(objDefn);
      return TCL_OK;
    case PROPERTY_OBJECT:
      /* We need to create the property in the object, but luckily
       * we've got the code to do this just above.
       */
      goto createAProperty;
    case PROPERTY_CLASS:
    default:
      Tcl_AppendResult(interp, "property \"", S(propName),
		       "\" is unwritable in object \"", S(objDefn->self),
		       "\"", NULL);
      return TCL_ERROR;
    }
  }
}

/* Implementation of an object - see documentation elsewhere */

static int
ObjImplCmd(clientData, interp, objc, objv)
     ClientData clientData;
     Tcl_Interp *interp;
     int objc;
     Tcl_Obj *CONST objv[];
{
  struct Object *objDefn = clientData;

  if (objc<2) {
    /* Default - return the class of the object */
    Tcl_IncrRefCount(objDefn->myClass->self);
    Tcl_SetObjResult(interp, objDefn->myClass->self);
    return TCL_OK;
  } else if (DereferenceP(interp, objv[1])) {
    /* Property access */
    if (objc<3) {
      Tcl_AppendResult(interp, "wrong # args: usage \"", S(objv[0]),
		       " -> propName ?operation ...?\"", NULL);
      return TCL_ERROR;
    } else {
      Tcl_HashEntry *hent;
      int defaulted;
      Tcl_Obj *value;

      if (objc==3) {
	/* Read a property */
	if (ReadObjectProp(interp, objDefn, objv[2], &value,
			   &defaulted, &hent) != TCL_OK) {
	  return TCL_ERROR;
	}
	Tcl_SetObjResult(interp, value);
	return TCL_OK;
      } else {
	/* Do something elaborate to a property */
	int i,index;
	long accumulator,addValue;
	Tcl_HashEntry *hashEnt;

	if (Tcl_GetIndexFromObj(interp, objv[3], propAccesses,
				"property access subcommand",
				0, &index)!=TCL_OK) {
	  return TCL_ERROR;
	}
	switch (index) {

	case 0: /* = - Write a property */
	  if (objc!=5) {
	    Tcl_AppendResult(interp, "wrong # args: should be \"", S(objv[0]),
			     " -> ", S(objv[2]), " = value\"", NULL);
	    return TCL_ERROR;
	  }
	  hashEnt = FindHashEntry(&objDefn->properties, objv[2]);
	  if (WriteObjectProp(interp, objDefn, objv[2], objv[4], 0,
			      hashEnt, &value) != TCL_OK) {
	    return TCL_ERROR;
	  }
	  Tcl_SetObjResult(interp, value);
	  return TCL_OK;

	case 1: /* append - Append strings to a property */
	  Tcl_Preserve(objDefn);
	  if (ReadObjectProp(interp, objDefn, objv[2], &value,
			     &defaulted, &hent) != TCL_OK) {
	    Tcl_Release(objDefn);
	    return TCL_ERROR;
	  }
	  value = Tcl_DuplicateObj(value);
	  for (i=4 ; i<objc ; i++) {
	    int length;
	    char *bytes = Tcl_GetStringFromObj(objv[i], &length);

	    Tcl_AppendToObj(value, bytes, length);
	  }
	  if (WriteObjectProp(interp, objDefn, objv[2], value, defaulted, hent,
			      &value) != TCL_OK) {
	    Tcl_Release(objDefn);
	    return TCL_ERROR;
	  }
	  Tcl_Release(objDefn);
	  Tcl_SetObjResult(interp, value);
	  return TCL_OK;

	case 2: /* exists - Check for existence of property */
	  if (objc!=4) {
	    Tcl_AppendResult(interp, "wrong # args: should be \"", S(objv[0]),
			     " -> ", S(objv[2]), " exists\"", NULL);
	    return TCL_ERROR;
	  }
	  hent = FindHashEntry(&objDefn->properties, objv[2]);
	  if (!hent) {
	    hent = FindHashEntry(&objDefn->myClass->properties, objv[2]);
	    if (hent) {
	      struct Property *propDefn = Tcl_GetHashValue(hent);
	      if (propDefn->kind == PROPERTY_CLASS)
		hent = NULL; /* Reading this would be an error! */
	    }
	  }
	  Tcl_SetObjResult(interp, Tcl_NewBooleanObj(hent!=NULL));
	  return TCL_OK;

	case 3: /* increment - Add to a property */
	  Tcl_Preserve(objDefn);
	  if (ReadObjectProp(interp, objDefn, objv[2], &value,
			     &defaulted, &hent) != TCL_OK) {
	    Tcl_Release(objDefn);
	    return TCL_ERROR;
	  }
	  accumulator = addValue = 0;
	  if (Tcl_GetLongFromObj(interp, value, &accumulator) != TCL_OK) {
	    Tcl_Release(objDefn);
	    return TCL_ERROR;
	  }
	  for (i=4 ; i<objc ; i++) {
	    if (Tcl_GetLongFromObj(interp, objv[i], &addValue) != TCL_OK) {
	      Tcl_Release(objDefn);
	      return TCL_ERROR;
	    }
	    accumulator += addValue;
	  }
	  value = Tcl_NewLongObj(accumulator);
	  Tcl_IncrRefCount(value);
	  if (WriteObjectProp(interp, objDefn, objv[2], value, defaulted, hent,
			      &value) != TCL_OK) {
	    Tcl_Release(objDefn);
	    Tcl_DecrRefCount(value);
	    return TCL_ERROR;
	  }
	  Tcl_Release(objDefn);
	  Tcl_SetObjResult(interp, value);
	  return TCL_OK;

	case 4: /* lappend - Append list items to a property */
	  Tcl_Preserve(objDefn);
	  if (ReadObjectProp(interp, objDefn, objv[2], &value,
			     &defaulted, &hent) != TCL_OK) {
	    Tcl_Release(objDefn);
	    return TCL_ERROR;
	  }
	  value = Tcl_DuplicateObj(value);
	  for (i=4 ; i<objc ; i++) {
	    if (Tcl_ListObjAppendElement(interp, value,
					 objv[i]) != TCL_OK) {
	      Tcl_DecrRefCount(value);
	      Tcl_Release(objDefn);
	      return TCL_ERROR;
	    }
	  }
	  if (WriteObjectProp(interp, objDefn, objv[2], value, defaulted, hent,
			      &value) != TCL_OK) {
	    Tcl_Release(objDefn);
	    return TCL_ERROR;
	  }
	  Tcl_Release(objDefn);
	  Tcl_SetObjResult(interp, value);
	  return TCL_OK;

	case 5: /* trace - Set/query a trace on a property */
	  if (objc<5 || objc>6) {
	    Tcl_AppendResult(interp, "wrong # args: should be \"", S(objv[0]),
			     " -> ", S(objv[2]), " trace read/write "
			     "?script?\"", NULL);
	    return TCL_ERROR;
	  } else {
	    struct Property *propDefn;
	    Tcl_Preserve(objDefn);
	    hent = FindHashEntry(&objDefn->properties, objv[2]);
	    if (!hent) {
	      hent = FindHashEntry(&objDefn->myClass->properties, objv[2]);
	      if (hent) {
		propDefn = Tcl_GetHashValue(hent);
		switch (propDefn->kind) {
		case PROPERTY_OBJECT:
		  /* Duplicate the property into the object */
		  WriteObjectProp(interp, objDefn, objv[2], propDefn->contents,
				  1, hent, NULL);
		  hent = FindHashEntry(&objDefn->properties, objv[2]);
		case PROPERTY_SHARED:
		  break;
		case PROPERTY_CLASS:
		default:
		  /* Reading this would be an error! */
		  Tcl_AppendResult(interp, "untraceable property \"",
				   S(objv[2]), "\" for object \"",
				   S(objv[0]), "\"", NULL);
		  Tcl_Release(objDefn);
		  return TCL_ERROR;
		}
	      } else {
		/* We're going to do something with a new object-only
		 * property.  This means that we have to create it.
		 */
		WriteObjectProp(interp, objDefn, objv[2], NewObj(),
				0, NULL, NULL);
		hent = FindHashEntry(&objDefn->properties, objv[2]);
	      }
	    }
	    propDefn = Tcl_GetHashValue(hent);
	    if (Tcl_GetIndexFromObj(interp, objv[4], traceCmds,
				    "property trace subcommand",
				    0, &index)!=TCL_OK) {
	      return TCL_ERROR;
	    }
	    switch (index) {
	    case 0: /* read */
	      if (objc==6) {
		Tcl_DecrRefCount(propDefn->trace_read_script);
		propDefn->trace_read_script = objv[5];
	      }
	      Tcl_IncrRefCount(propDefn->trace_read_script);
	      Tcl_SetObjResult(interp, propDefn->trace_read_script);
	      Tcl_Release(objDefn);
	      return TCL_OK;

	    case 1: /* write */
	      if (objc==6) {
		Tcl_DecrRefCount(propDefn->trace_write_script);
		propDefn->trace_write_script = objv[5];
	      }
	      Tcl_IncrRefCount(propDefn->trace_write_script);
	      Tcl_SetObjResult(interp, propDefn->trace_write_script);
	      Tcl_Release(objDefn);
	      return TCL_OK;

	    default:
	      Tcl_AppendResult(interp, "switch fall-through", NULL);
	      Tcl_Release(objDefn);
	      return TCL_ERROR;
	    }
	  }

	case 6: /* unset - Delete a property */
	  if (objc!=4) {
	    Tcl_AppendResult(interp, "wrong # args: should be \"", S(objv[0]),
			     " -> ", S(objv[2]), " unset\"", NULL);
	  } else {
	    struct Property *propDefn;
	    hent = FindHashEntry(&objDefn->properties, objv[2]);

	    if (hent) {
	      propDefn = Tcl_GetHashValue(hent);
	      Tcl_EventuallyFree(propDefn, DelPropForReal);
	      Tcl_DeleteHashEntry(hent);
	      return TCL_OK;
	    } else {
	      hent = FindHashEntry(&objDefn->myClass->properties, objv[2]);
	      if (hent) {
		propDefn = Tcl_GetHashValue(hent);
		if (propDefn->kind == PROPERTY_SHARED) {
		  /* Set to the empty string */
		  Tcl_Preserve(objDefn);
		  value = Tcl_NewObj();
		  if (WriteObjectProp(interp, objDefn, objv[2], value,
				      0, hent, NULL) != TCL_OK) {
		    Tcl_DecrRefCount(value);
		    Tcl_Release(objDefn);
		    return TCL_ERROR;
		  }
		  Tcl_Release(objDefn);
		  return TCL_OK;
		}
	      }
	    }
	    Tcl_AppendResult(interp, "unknown property \"", S(objv[2]),
			     "\" for object \"", S(objv[0]), "\"", NULL);
	    return TCL_ERROR;
	  }

	case 7: /* link - mirror a property in a global variable */
	  if (objc<4 || objc>5) {
	    Tcl_AppendResult(interp, "wrong # args: should be \"", S(objv[0]),
			     " -> ", S(objv[2]), " link ?variable?\"", NULL);
	  } else {
	    struct Property *propDefn;
	    Tcl_Preserve(objDefn);
	    hent = FindHashEntry(&objDefn->properties, objv[2]);
	    if (!hent) {
	      hent = FindHashEntry(&objDefn->myClass->properties, objv[2]);
	      if (hent) {
		propDefn = Tcl_GetHashValue(hent);
		switch (propDefn->kind) {
		case PROPERTY_OBJECT:
		  /* Duplicate the property into the object */
		  WriteObjectProp(interp, objDefn, objv[2], propDefn->contents,
				  1, hent, NULL);
		  hent = FindHashEntry(&objDefn->properties, objv[2]);
		case PROPERTY_SHARED:
		case PROPERTY_CLASS:
		  break;
		default:
		  /* Reading this would be an error! */
		  Tcl_AppendResult(interp, "unlinkable property \"",
				   S(objv[2]), "\" for object \"",
				   S(objv[0]), "\"", NULL);
		  Tcl_Release(objDefn);
		  return TCL_ERROR;
		}
	      } else {
		/* We're going to do something with a new object-only
		 * property.  This means that we have to create it.
		 */
		WriteObjectProp(interp, objDefn, objv[2], Tcl_NewObj(),
				0, NULL, NULL);
		hent = FindHashEntry(&objDefn->properties, objv[2]);
	      }
	    }
	    propDefn = Tcl_GetHashValue(hent);
	    if (objc==5) {
	      int vlen;

	      Tcl_GetStringFromObj(objv[4], &vlen);

	      if (vlen) {
		if (Dkfobj_LinkVar(interp, objv[4], &propDefn->contents,
				   propDefn->kind == PROPERTY_CLASS)!=TCL_OK) {
		  Tcl_Release(objDefn);
		  return TCL_ERROR;
		}
	      }
	      if (propDefn->is_linked) {
		Dkfobj_UnlinkVar(interp, propDefn->linked_var);
		propDefn->is_linked = 0;
	      }
	      Tcl_DecrRefCount(propDefn->linked_var);
	      propDefn->linked_var = objv[4];
	      Tcl_IncrRefCount(propDefn->linked_var);
	      if (vlen) {
		propDefn->is_linked = 1;
	      }
	    }
	    Tcl_SetObjResult(interp, propDefn->linked_var);
	    Tcl_Release(objDefn);
	    return TCL_OK;
	  }

	default:
	  Tcl_AppendResult(interp, "switch fall-though", NULL);
	  return TCL_ERROR;
	}
      }
    }
  } else {
    /* Method call */
    Tcl_HashEntry *hent =
      FindHashEntry(&objDefn->myClass->methods, objv[1]);

    if (!hent) {
      Tcl_AppendResult(interp, "unknown method \"", S(objv[1]),
		       "\" for object \"", S(objv[0]), "\"", NULL);
      return TCL_ERROR;
    } else {
      struct Method *mthdDefn = Tcl_GetHashValue(hent);
      int code;
      switch (mthdDefn->kind) {
      case METHOD_OBJECT:
      case METHOD_SHARED:
	code = MethodCall(interp, mthdDefn->commandName, objDefn->self,
			  objc-2, objv+2);
	if (code != TCL_OK) {
	  Tcl_AddObjErrorInfo(interp, "\n    (during object method call)", -1);
	}
	return code;
      case METHOD_CONSTRUCTOR:
      case METHOD_CLASS:
      default: /* Shouldn't happen */
	Tcl_AppendResult(interp, "method \"", S(objv[1]),
			 "\" is not legally callable for object \"",
			 S(objDefn->self), "\"", NULL);
	return TCL_ERROR;
      }
    }
  }
}

/* ---------------------------------------------------------------------- */

/* Parses a definition list, adding bits to the class being created */

static int
ClassAddDefinition(interp, newClass, defList, classes)
     Tcl_Interp *interp;
     struct Class *newClass;
     Tcl_Obj *defList;
     Tcl_HashTable *classes;
{
  Tcl_HashSearch hsearch;
  Tcl_HashEntry *hent;
  int i,deflen,isNew;
  enum MethKind methkind = METHOD_SHARED;
  enum PropKind propkind = PROPERTY_OBJECT;

  if (Tcl_ListObjLength(interp, defList, &deflen) != TCL_OK) {
    return TCL_ERROR;
  }

  for (i=0 ; i<deflen ; i++) {
    Tcl_Obj *defnObj;
    int index = Tcl_ListObjIndex(interp, defList, i, &defnObj);
    char *defn = Tcl_GetStringFromObj(defnObj, NULL);
    struct Property *propDefn;
    struct Method *mthdDefn;
    struct Class *clsDefn;

    if (Tcl_GetIndexFromObj(interp, defnObj, defnkeywords, "keyword",
			    0, &index)!=TCL_OK) {
      return TCL_ERROR;
    }
    switch (index) {
    case 0: /* class: */
      methkind = METHOD_CLASS;
      propkind = PROPERTY_CLASS;
      break;

    case 1: /* constructor: */
      methkind = METHOD_CONSTRUCTOR;
      break;

    case 2: /* inherit */
      if (i+1>=deflen) {
	Tcl_AppendResult(interp, "bad inherit declaration: should be \"",
			 S(defnObj), " className\"", NULL);
	return TCL_ERROR;
      }
      Tcl_ListObjIndex(interp, defList, i+1, &defnObj);
      hent = FindHashEntry(classes, defnObj);
      if (!hent) {
	Tcl_AppendResult(interp, "unknown class \"", S(defnObj), "\"", NULL);
	return TCL_ERROR;
      }
      Tcl_ListObjAppendElement(interp, newClass->parents, defnObj);
      clsDefn = Tcl_GetHashValue(hent);
      /* Duplicate all the methods */
      hent = Tcl_FirstHashEntry(&clsDefn->methods, &hsearch);
      for (;hent; hent=Tcl_NextHashEntry(&hsearch)) {
	Tcl_HashEntry *nhent;
	struct Method *parentM = Tcl_GetHashValue(hent);
	char *key = Tcl_GetStringFromObj(parentM->name, NULL);
	struct Method *ourM = (struct Method *)
	  Tcl_Alloc(sizeof(struct Method));

	ourM->kind = parentM->kind;
	ourM->name = parentM->name;
	ourM->commandName = parentM->commandName;
	Tcl_IncrRefCount(ourM->name);
	Tcl_IncrRefCount(ourM->commandName);

	nhent = Tcl_CreateHashEntry(&newClass->methods, key, &isNew);
	if (!isNew) {
	  Tcl_EventuallyFree(Tcl_GetHashValue(nhent), DelMethodForReal);
	}
	Tcl_SetHashValue(nhent,ourM);
      }
      /* Duplicate all the properties */
      hent = Tcl_FirstHashEntry(&clsDefn->properties, &hsearch);
      for (;hent; hent=Tcl_NextHashEntry(&hsearch)) {
	Tcl_HashEntry *nhent;
	struct Property *parentP = Tcl_GetHashValue(hent);
	char *key = Tcl_GetHashKey(&clsDefn->properties, hent);
	Tcl_Obj *writer = NewObj();
	struct Property *ourP =
	  MakeProperty(parentP->kind, parentP->name, parentP->contents,
		       writer, interp);

	Tcl_DecrRefCount(writer);
	nhent = Tcl_CreateHashEntry(&newClass->properties, key, &isNew);
	if (!isNew) {
	  Tcl_EventuallyFree(Tcl_GetHashValue(nhent), DelPropForReal);
	}
	Tcl_SetHashValue(nhent,ourP);
      }
      i++;
      break;

    case 3: /* method */
      if (i+2>=deflen) {
	Tcl_AppendResult(interp, "bad method declaration: should be \"",
			 S(defnObj), " methodName methodProcName\"", NULL);
	return TCL_ERROR;
      }
      mthdDefn = (struct Method *)Tcl_Alloc(sizeof(struct Method));
      mthdDefn->kind = methkind;
      /* The next two array lookups shouldn't fail... */
      Tcl_ListObjIndex(interp, defList, i+1, &mthdDefn->name);
      Tcl_ListObjIndex(interp, defList, i+2, &mthdDefn->commandName);
      Tcl_IncrRefCount(mthdDefn->name);
      Tcl_IncrRefCount(mthdDefn->commandName);
      defn = Tcl_GetStringFromObj(mthdDefn->name, NULL);
      hent = Tcl_CreateHashEntry(&newClass->methods, defn, &isNew);
      if (!isNew) {
	Tcl_EventuallyFree(Tcl_GetHashValue(hent), DelMethodForReal);
      }
      Tcl_SetHashValue(hent,mthdDefn);
      i+=2;
      break;

    case 4: /* object: */
      methkind = METHOD_OBJECT;
      propkind = PROPERTY_OBJECT;
      break;

    case 5: /* property */
      if (i+2>=deflen) {
	Tcl_AppendResult(interp, "bad property declaration: should be \"",
			 S(defnObj), " propertyName initialValue\"", NULL);
	return TCL_ERROR;
      } else {
	Tcl_Obj *Pname, *Pcontents, *Pwriter = NewObj();

	/* The next two list operations shouldn't fail... */
	Tcl_ListObjIndex(interp, defList, i+1, &Pname);
	Tcl_ListObjIndex(interp, defList, i+2, &Pcontents);

	propDefn = MakeProperty(propkind, Pname, Pcontents, Pwriter, interp);
	Tcl_DecrRefCount(Pwriter);
	defn = Tcl_GetStringFromObj(Pname, NULL);
      }
      hent = Tcl_CreateHashEntry(&newClass->properties, defn, &isNew);
      if (!isNew) {
	Tcl_EventuallyFree(Tcl_GetHashValue(hent), DelPropForReal);
      }
      Tcl_SetHashValue(hent,propDefn);
      i+=2;
      break;

    case 6: /* shared: */
      methkind = METHOD_SHARED;
      propkind = PROPERTY_SHARED;
      break;

    case 7: /* virtual */
      newClass->virtual = 1;
      break;

    default: /* This is probably unreachable */
      Tcl_AppendResult(interp, "unhandled case \"", S(defnObj), "\"", NULL);
      return TCL_ERROR;
    }
  }
  return TCL_OK;
}

/* Implementation of the class system command - see documentation elsewhere */

static int
ClassCommand(clientData, interp, objc, objv)
     ClientData clientData;
     Tcl_Interp *interp;
     int objc;
     Tcl_Obj *CONST objv[];
{
  Tcl_HashTable *classes = (Tcl_HashTable *)clientData;
  Tcl_HashSearch hsearch;
  Tcl_HashEntry *hent;
  int len, index;

  if (objc<2) {
    Tcl_AppendResult(interp, "wrong # args: should be \"", S(objv[0]),
		     " command ?arg ...?\"", NULL);
    return TCL_ERROR;
  }
  if (Tcl_GetIndexFromObj(interp, objv[1], classcmds, "subcommand",
			  0, &index)!=TCL_OK) {
    return TCL_ERROR;
  }
  switch (index) {
  case 0: /* define - Define a new class! */ {
    struct Class *newClass;
    int i;
    char *name;

    if (objc<4) {
      Tcl_AppendResult(interp, "wrong # args: should be \"", S(objv[0]),
		       " define name deflist ?deflist ...?\"", NULL);
      return TCL_ERROR;
    }

    for (i=3 ; i<objc ; i++) {
      /* Verify that we have a list... */
      if (Tcl_ListObjLength(interp, objv[i], &len) != TCL_OK) {
	return TCL_ERROR;
      }
    }

    newClass = (struct Class *)Tcl_Alloc(sizeof(struct Class));
    Tcl_InitHashTable(&newClass->methods,    TCL_STRING_KEYS);
    Tcl_InitHashTable(&newClass->properties, TCL_STRING_KEYS);
    Tcl_InitHashTable(&newClass->objects,    TCL_STRING_KEYS);
    newClass->parents = NewObj();
    newClass->creatorInterp = interp;
    newClass->virtual = 0;

    for (i=3 ; i<objc ; i++) {
      /* Build the class... */
      if (ClassAddDefinition(interp,newClass,objv[i],classes) != TCL_OK) {
	Tcl_DecrRefCount(newClass->parents);
	/* Need to delete the hash tables properly */
	Tcl_Free((char *)newClass);
	return TCL_ERROR;
      }
    }
    newClass->self = objv[2];
    Tcl_IncrRefCount(objv[2]);
    name = Tcl_GetStringFromObj(objv[2], &len);
    newClass->classes = classes;
    newClass->token =
      Tcl_CreateObjCommand(interp, name, ClassImplCmd,
			   newClass, DelClass);
    Tcl_SetHashValue(Tcl_CreateHashEntry(classes, name, &len), newClass);
    Tcl_SetObjResult(interp, newClass->self);
    return TCL_OK;
  }

  case 1: /* exists - Do we have a class name? */
    if (objc!=3) {
      Tcl_AppendResult(interp, "wrong # args: should be \"", S(objv[0]),
		       " exists name\"", NULL);
      return TCL_ERROR;
    }
    hent = FindHashEntry(classes, objv[2]);
    Tcl_SetObjResult(interp, Tcl_NewIntObj(hent != NULL));
    return TCL_OK;

  case 2: /* info - Find something out about something specific */ {
    struct Object *obj;
    struct Class *cls;
    int inObj;

    if (objc<3) {
      Tcl_AppendResult(interp, "wrong # args: should be \"", S(objv[0]),
		       " info thing arg ?arg?\"", NULL);
      return TCL_ERROR;
    }

    if (Tcl_GetIndexFromObj(interp, objv[2], classinfos,
			    "kind of information", 0, &index)!=TCL_OK) {
      return TCL_ERROR;
    }

    switch (index) {
    case 0: /* class */
      if (objc!=4) {
	Tcl_AppendResult(interp, "wrong # args: should be \"", S(objv[0]),
			 " info class className\"", NULL);
	return TCL_ERROR;
      }

      hent = FindHashEntry(classes, objv[3]);
      if (hent) {
	Tcl_Obj *ret = Tcl_NewListObj(0,NULL);
	cls = (struct Class *)Tcl_GetHashValue(hent);
	Tcl_ListObjAppendElement(interp, ret, Tcl_NewIntObj(cls->virtual));
	Tcl_ListObjAppendElement(interp, ret, cls->parents);
	Tcl_SetObjResult(interp, ret);
	return TCL_OK;
      } else {
	Tcl_AppendResult(interp, "unknown class \"", S(objv[3]), "\"", NULL);
	return TCL_ERROR;
      }

    case 2: /* object */
      if (objc!=4) {
	Tcl_AppendResult(interp, "wrong # args: should be \"", S(objv[0]),
			 " info object objectName\"", NULL);
	return TCL_ERROR;
      }

      if (GetObject(interp, objv[3], classes, &obj) != TCL_OK) {
	return TCL_ERROR;
      } else {
	Tcl_Obj *ret = Tcl_NewListObj(0,NULL);

	Tcl_ListObjAppendElement(interp, ret, obj->myClass->self);
	Tcl_ListObjAppendElement(interp, ret, obj->built_with_constr);
	Tcl_SetObjResult(interp, ret);
	return TCL_OK;
      }

    case 1: /* method */
      if (objc!=5) {
	Tcl_AppendResult(interp, "wrong # args: should be \"", S(objv[0]),
			 " info method classOrObject methodName\"", NULL);
	return TCL_ERROR;
      }

      hent = FindHashEntry(classes, objv[3]);
      if (hent) {
	cls = (struct Class *)Tcl_GetHashValue(hent);
	inObj = 0;
      } else {
	if (GetObject(interp, objv[3], classes, &obj) != TCL_OK) {
	  Tcl_AppendResult(interp, "\"", S(objv[3]),
			   "\" is neither a class or an object", NULL);
	  return TCL_ERROR;
	}
	cls = obj->myClass;
	inObj = 1;
      }
      hent = FindHashEntry(&cls->methods, objv[4]);
      if (hent) {
	struct Method *mthdDefn = (struct Method *)Tcl_GetHashValue(hent);
	Tcl_Obj *ret;

	if ((inObj  && (mthdDefn->kind==METHOD_CONSTRUCTOR||
			mthdDefn->kind==METHOD_CLASS)) ||
	    (!inObj && (mthdDefn->kind==METHOD_OBJECT)))
	  goto unknownMethodForInfo;
	ret = Tcl_NewListObj(0,NULL);
	switch (mthdDefn->kind) {
	case METHOD_CONSTRUCTOR:
	  Tcl_ListObjAppendElement(interp, ret,
				   Tcl_NewStringObj("constructor",-1));
	  break;
	case METHOD_CLASS:
	  Tcl_ListObjAppendElement(interp, ret, Tcl_NewStringObj("class",-1));
	  break;
	case METHOD_OBJECT:
	  Tcl_ListObjAppendElement(interp, ret, Tcl_NewStringObj("object",-1));
	  break;
	case METHOD_SHARED:
	  Tcl_ListObjAppendElement(interp, ret, Tcl_NewStringObj("shared",-1));
	  break;
	default:
	  Tcl_ListObjAppendElement(interp, ret, Tcl_NewStringObj("???",-1));
	  break;
	}
	Tcl_ListObjAppendElement(interp, ret, mthdDefn->commandName);
	Tcl_SetObjResult(interp, ret);
	return TCL_OK;
      } else {
      unknownMethodForInfo:
	Tcl_AppendResult(interp, "unknown method \"", S(objv[4]), "\"", NULL);
	return TCL_ERROR;
      }

    case 3: /* property */
      if (objc!=5) {
	Tcl_AppendResult(interp, "wrong # args: should be \"", S(objv[0]),
			 " info property classOrObject propertyName\"", NULL);
	return TCL_ERROR;
      }

      hent = FindHashEntry(classes, objv[3]);
      if (hent) {
	cls = (struct Class *)Tcl_GetHashValue(hent);
	inObj = 0;
	hent = FindHashEntry(&cls->properties, objv[4]);
      } else {
	if (GetObject(interp, objv[3], classes, &obj) != TCL_OK) {
	  Tcl_AppendResult(interp, "\"", S(objv[3]),
			   "\" is neither a class or an object", NULL);
	  return TCL_ERROR;
	}
	inObj = 1;
	hent = FindHashEntry(&obj->properties, objv[4]);
      }
      if (hent) {
	struct Property *propDefn = (struct Property *)Tcl_GetHashValue(hent);
	Tcl_Obj *ret;

	if (inObj && propDefn->kind==PROPERTY_CLASS)
	  goto unknownPropertyForInfo;
	ret = Tcl_NewListObj(0,NULL);
	switch (propDefn->kind) {
	case PROPERTY_CLASS:
	  Tcl_ListObjAppendElement(interp, ret, Tcl_NewStringObj("class",-1));
	  break;
	case PROPERTY_OBJECT:
	  Tcl_ListObjAppendElement(interp, ret, Tcl_NewStringObj("object",-1));
	  break;
	case PROPERTY_SHARED:
	  Tcl_ListObjAppendElement(interp, ret, Tcl_NewStringObj("shared",-1));
	  break;
	default:
	  Tcl_ListObjAppendElement(interp, ret, Tcl_NewStringObj("???",-1));
	  break;
	}
	Tcl_ListObjAppendElement(interp, ret, propDefn->writer);
	Tcl_SetObjResult(interp, ret);
	return TCL_OK;
      } else {
      unknownPropertyForInfo:
	Tcl_AppendResult(interp, "unknown property \"", S(objv[4]), "\"",
			 NULL);
	return TCL_ERROR;
      }
    }
    break;
  }

  case 3: /* list - Lists of things */ {
    Tcl_Obj *list;
    if (objc<3) {
      Tcl_AppendResult(interp, "wrong # args: should be \"",
		       S(objv[0]), " list thing ?arg ...?\"", NULL);
      return TCL_ERROR;
    }
    if (Tcl_GetIndexFromObj(interp, objv[2], classlists,
			    "kind of list", 0, &index)!=TCL_OK) {
      return TCL_ERROR;
    }

    switch (index) {
    case 0: /* classes */ {
      char *pattern;

      switch (objc) {
      case 4:
	pattern = Tcl_GetStringFromObj(objv[3], NULL);
	list = Tcl_NewListObj(0,NULL);
	hent = Tcl_FirstHashEntry(classes,&hsearch);
	for (;hent; hent=Tcl_NextHashEntry(&hsearch)) {
	  Tcl_Obj *cls = ((struct Class *)Tcl_GetHashValue(hent))->self;
	  if (Tcl_StringMatch(Tcl_GetStringFromObj(cls,NULL), pattern))
	    Tcl_ListObjAppendElement(interp, list, cls);
	}
	Tcl_SetObjResult(interp, list);
	return TCL_OK;
      case 3:
	list = Tcl_NewListObj(0,NULL);
	hent = Tcl_FirstHashEntry(classes,&hsearch);
	for (;hent; hent=Tcl_NextHashEntry(&hsearch)) {
	  Tcl_ListObjAppendElement(interp, list, ((struct Class *)
				   Tcl_GetHashValue(hent))->self);
	}
	Tcl_SetObjResult(interp, list);
	return TCL_OK;
      default:
	Tcl_AppendResult(interp, "wrong # args: should be \"", S(objv[0]),
			 " list classes ?pattern?\"", NULL);
	return TCL_ERROR;
      }
    }

    case 1: /* methods */ {
      char *pattern;
      struct Class *cls;
      int inObj;

      switch (objc) {
      case 4:
	hent = FindHashEntry(classes, objv[3]);
	if (hent) {
	  cls = (struct Class *)Tcl_GetHashValue(hent);
	  inObj = 0;
	} else {
	  struct Object *obj;
	  if (GetObject(interp, objv[3], classes, &obj) != TCL_OK) {
	    Tcl_AppendResult(interp, "\"", S(objv[3]),
			     "\" is neither a class or an object", NULL);
	    return TCL_ERROR;
	  }
	  cls = obj->myClass;
	  inObj = 1;
	}
	hent = Tcl_FirstHashEntry(&cls->methods, &hsearch);
	list = Tcl_NewListObj(0,NULL);
	for (;hent; hent=Tcl_NextHashEntry(&hsearch)) {
	  struct Method *mthdDefn = (struct Method *)Tcl_GetHashValue(hent);
	  if ((inObj  && (mthdDefn->kind==METHOD_OBJECT ||
			  mthdDefn->kind==METHOD_SHARED)) ||
	      (!inObj && mthdDefn->kind!=METHOD_OBJECT))
	    Tcl_ListObjAppendElement(interp, list, mthdDefn->name);
	}
	Tcl_SetObjResult(interp, list);
	return TCL_OK;

      case 5:
	pattern = Tcl_GetStringFromObj(objv[4], &len);
	hent = FindHashEntry(classes, objv[3]);
	if (hent) {
	  cls = (struct Class *)Tcl_GetHashValue(hent);
	  inObj = 0;
	} else {
	  struct Object *obj;
	  if (GetObject(interp, objv[3], classes, &obj) != TCL_OK) {
	    Tcl_AppendResult(interp, "\"", S(objv[3]),
			     "\" is neither a class or an object", NULL);
	    return TCL_ERROR;
	  }
	  cls = obj->myClass;
	  inObj = 1;
	}
	hent = Tcl_FirstHashEntry(&cls->methods, &hsearch);
	list = Tcl_NewListObj(0,NULL);
	for (;hent; hent=Tcl_NextHashEntry(&hsearch)) {
	  struct Method *mthdDefn = (struct Method *)Tcl_GetHashValue(hent);
	  if ((inObj &&  (mthdDefn->kind==2 || mthdDefn->kind==3)) ||
	      (!inObj && mthdDefn->kind!=2)) {
	    char *name = Tcl_GetStringFromObj(mthdDefn->name, NULL);
	    if (Tcl_StringMatch(name, pattern))
	      Tcl_ListObjAppendElement(interp, list, mthdDefn->name);
	  }
	}
	Tcl_SetObjResult(interp, list);
	return TCL_OK;

      default:
	Tcl_AppendResult(interp, "wrong # args: should be \"", S(objv[0]),
			 " list methods classOrObject ?pattern?\"", NULL);
	return TCL_ERROR;
      }
    }

    case 2: /* objects */ {
      char *pattern;
      struct Class *cls;

      switch (objc) {
      case 5:
	hent = FindHashEntry(classes, objv[3]);
	if (hent) {
	  cls = (struct Class *)Tcl_GetHashValue(hent);
	} else {
	  Tcl_AppendResult(interp, "unknown class \"", S(objv[3]), "\"", NULL);
	  return TCL_ERROR;
	}
	pattern = Tcl_GetStringFromObj(objv[4], NULL);
	list = Tcl_NewListObj(0,NULL);
	hent = Tcl_FirstHashEntry(&cls->objects,&hsearch);
	for (;hent; hent=Tcl_NextHashEntry(&hsearch)) {
	  Tcl_Obj *obj = ((struct Object *)Tcl_GetHashValue(hent))->self;
	  if (Tcl_StringMatch(Tcl_GetStringFromObj(obj, NULL), pattern))
	    Tcl_ListObjAppendElement(interp, list, obj);
	}
	Tcl_SetObjResult(interp, list);
	return TCL_OK;

      case 4:
	hent = FindHashEntry(classes, objv[3]);
	if (hent) {
	  cls = (struct Class *)Tcl_GetHashValue(hent);
	} else {
	  Tcl_AppendResult(interp, "unknown class \"", S(objv[3]), "\"", NULL);
	  return TCL_ERROR;
	}
	list = Tcl_NewListObj(0,NULL);
	hent = Tcl_FirstHashEntry(&cls->objects,&hsearch);
	for (;hent; hent=Tcl_NextHashEntry(&hsearch)) {
	  Tcl_ListObjAppendElement(interp,list,((struct Object *)
						Tcl_GetHashValue(hent))->self);
	}
	Tcl_SetObjResult(interp, list);
	return TCL_OK;

      default:
	Tcl_AppendResult(interp, "wrong # args: should be \"", S(objv[0]),
			 " list objects class ?pattern?\"", NULL);
	return TCL_ERROR;
      }
    }

    case 3: /* properties */ {
      char *pattern;
      struct Class *cls;
      int inObj;

      switch (objc) {
      case 4:
	hent = FindHashEntry(classes, objv[3]);
	if (hent) {
	  cls = (struct Class *)Tcl_GetHashValue(hent);
	  inObj = 0;
	  hent = Tcl_FirstHashEntry(&cls->properties, &hsearch);
	} else {
	  struct Object *obj;
	  if (GetObject(interp, objv[3], classes, &obj) != TCL_OK) {
	    Tcl_AppendResult(interp, "\"", S(objv[3]),
			     "\" is neither a class or an object", NULL);
	    return TCL_ERROR;
	  }
	  cls = obj->myClass;
	  inObj = 1;
	  hent = Tcl_FirstHashEntry(&obj->properties, &hsearch);
	}
	list = Tcl_NewListObj(0,NULL);
	for (;hent; hent=Tcl_NextHashEntry(&hsearch)) {
	  struct Property *propDefn = Tcl_GetHashValue(hent);
	  if (!inObj || propDefn->kind!=0)
	    Tcl_ListObjAppendElement(interp, list, propDefn->name);
	}
	Tcl_SetObjResult(interp, list);
	return TCL_OK;

      case 5:
	pattern = Tcl_GetStringFromObj(objv[4], &len);
	hent = FindHashEntry(classes, objv[3]);
	if (hent) {
	  cls = (struct Class *)Tcl_GetHashValue(hent);
	  inObj = 0;
	  hent = Tcl_FirstHashEntry(&cls->properties, &hsearch);
	} else {
	  struct Object *obj;
	  if (GetObject(interp, objv[3], classes, &obj) != TCL_OK) {
	    Tcl_AppendResult(interp, "\"", S(objv[3]),
			     "\" is neither a class or an object", NULL);
	    return TCL_ERROR;
	  }
	  cls = obj->myClass;
	  inObj = 1;
	  hent = Tcl_FirstHashEntry(&obj->properties, &hsearch);
	}
	list = Tcl_NewListObj(0,NULL);
	for (;hent; hent=Tcl_NextHashEntry(&hsearch)) {
	  struct Property *propDefn = Tcl_GetHashValue(hent);
	  if (!inObj || propDefn->kind!=0) {
	    char *name = Tcl_GetStringFromObj(propDefn->name, NULL);
	    if (Tcl_StringMatch(name, pattern))
	      Tcl_ListObjAppendElement(interp, list, propDefn->name);
	  }
	}
	Tcl_SetObjResult(interp, list);
	return TCL_OK;

      default:
	Tcl_AppendResult(interp, "wrong # args: should be \"", S(objv[0]),
			 " list methods classOrObject ?pattern?\"", NULL);
	return TCL_ERROR;
      }
    }
    }
    break;
  }

  case 4: /* of - What is the class of an object? */ {
    struct Object *objDefn;
    if (objc!=3) {
      Tcl_AppendResult(interp, "wrong # args: should be \"", S(objv[1]),
		       " of object\"", NULL);
      return TCL_ERROR;
    }
    if (GetObject(interp, objv[2], classes, &objDefn) != TCL_OK) {
      return TCL_ERROR;
    }
    Tcl_SetObjResult(interp, objDefn->myClass->self);
    return TCL_OK;
  }
  }
  Tcl_AppendResult(interp, "switch fall-through", NULL);
  return TCL_ERROR;
}

/* ---------------------------------------------------------------------- */

/* Deletion commands.
 *
 * Everything with the 'ForReal' suffix is called as a of result
 * Tcl_EventuallyFree (and is thus part of something which has been
 * Tcl_Preserve-d), whereas the other deletion commands are called as
 * part of deleting Tcl commands
 */

static void
DelMethodForReal(clientData)
     char *clientData;
{
  struct Method *methDefn = (struct Method *)clientData;
  Tcl_DecrRefCount(methDefn->name);
  Tcl_DecrRefCount(methDefn->commandName);
  Tcl_Free((char *)methDefn);
}

static void
DelPropForReal(clientData)
     char *clientData;
{
  struct Property *propDefn = (struct Property *)clientData;
  if (propDefn->is_linked) {
    Dkfobj_UnlinkVar(propDefn->interp, propDefn->linked_var);
  }
  Tcl_DecrRefCount(propDefn->name);
  Tcl_DecrRefCount(propDefn->contents);
  Tcl_DecrRefCount(propDefn->writer);
  Tcl_DecrRefCount(propDefn->trace_write_script);
  Tcl_DecrRefCount(propDefn->trace_read_script);
  Tcl_DecrRefCount(propDefn->linked_var);
  Tcl_Free((char *)propDefn);
}

static void
DelObjForReal(clientData)
     char *clientData;
{
  struct Object *objDefn = (struct Object *)clientData;
  Tcl_HashEntry *hent;
  Tcl_HashSearch hsearch;

  hent = Tcl_FirstHashEntry(&objDefn->properties, &hsearch);
  for (;hent; hent=Tcl_NextHashEntry(&hsearch)) {
    Tcl_EventuallyFree(Tcl_GetHashValue(hent), DelPropForReal);
  }
  Tcl_DeleteHashTable(&objDefn->properties);

  Tcl_DecrRefCount(objDefn->self);
  Tcl_DecrRefCount(objDefn->built_with_constr);
  Tcl_Free((char *)objDefn);
}

static void
DelClassForReal(clientData)
     char *clientData;
{
  struct Class *myClass = (struct Class *)clientData;
  Tcl_HashSearch hsearch;
  Tcl_HashEntry *hent;

  hent = Tcl_FirstHashEntry(&myClass->objects, &hsearch);
  if (myClass->creatorInterp && !Tcl_InterpDeleted(myClass->creatorInterp)) {
    for (;hent; hent = Tcl_FirstHashEntry(&myClass->objects, &hsearch)) {
      struct Object *obj = Tcl_GetHashValue(hent);
      Tcl_DeleteCommand(myClass->creatorInterp,
			Tcl_GetCommandName(myClass->creatorInterp,
					   obj->token));
    }
  } else {
    for (;hent; hent = Tcl_NextHashEntry(&hsearch)) {
      struct Object *obj = Tcl_GetHashValue(hent);
      obj->myClass = NULL;
    }
  }
  Tcl_DeleteHashTable(&myClass->objects);

  hent = Tcl_FirstHashEntry(&myClass->methods, &hsearch);
  for (;hent; hent = Tcl_NextHashEntry(&hsearch)) {
    Tcl_EventuallyFree(Tcl_GetHashValue(hent), DelMethodForReal);
  }
  Tcl_DeleteHashTable(&myClass->methods);

  hent = Tcl_FirstHashEntry(&myClass->properties, &hsearch);
  for (;hent; hent = Tcl_NextHashEntry(&hsearch)) {
    Tcl_EventuallyFree(Tcl_GetHashValue(hent), DelPropForReal);
  }
  Tcl_DeleteHashTable(&myClass->properties);

  Tcl_DecrRefCount(myClass->parents);
  Tcl_DecrRefCount(myClass->self);
}

static void
DelObjCmd(clientData)
     ClientData clientData;
{
  struct Object *objDefn = (struct Object *)clientData;
  Tcl_HashSearch hsearch;
  Tcl_HashEntry *hent;

  if (objDefn->myClass) {
    hent = FindHashEntry(&objDefn->myClass->objects, objDefn->self);
    if (hent) Tcl_DeleteHashEntry(hent);
    objDefn->myClass = NULL;
  }

  /* Need to invalidate all references to this object */
  hent = Tcl_FirstHashEntry(&objDefn->references, &hsearch);
  for (;hent; hent=Tcl_NextHashEntry(&hsearch)) {
    Tcl_Obj *objPtr = (Tcl_Obj *)Tcl_GetHashKey(&objDefn->references, hent);
    objPtr->typePtr = NULL;
  }
  /* We can delete the hash table now, since the object is no longer
   * registered with the class - the only route available to the
   * creation of new references... */
  Tcl_DeleteHashTable(&objDefn->references);

  Tcl_EventuallyFree(clientData, DelObjForReal);
}

static void
DelClass(clientData)
     ClientData clientData;
{
  struct Class *myClass = (struct Class *)clientData;
  Tcl_HashEntry *hent;

  /* Stop further examination of this class, which is now officially
   * on its way out... */
  if (myClass->classes) {
    hent = FindHashEntry(myClass->classes, myClass->self);
    if (hent) Tcl_DeleteHashEntry(hent);
    myClass->classes = NULL;
  }

  /* Do the rest when it is known to be safe... */
  Tcl_EventuallyFree(clientData, DelClassForReal);
}

static void
DelClassCmd(clientData)
     ClientData clientData;
{
  Tcl_HashTable *classes = (Tcl_HashTable *)clientData;
  Tcl_HashSearch hsearch;
  Tcl_HashEntry *hent;
  Tcl_Interp *interp = NULL;

  hent = Tcl_FirstHashEntry(classes, &hsearch);
  if (hent) {
    struct Class *myClass = Tcl_GetHashValue(hent);
    if (!Tcl_InterpDeleted(myClass->creatorInterp))
      interp = myClass->creatorInterp;
  }
  for (;hent; hent = Tcl_FirstHashEntry(classes, &hsearch)) {
    struct Class *myClass = Tcl_GetHashValue(hent);
    if (interp) {
      Tcl_DeleteCommand(interp, Tcl_GetCommandName(interp, myClass->token));
    } else {
      myClass->classes = NULL;
    }
  }
  Tcl_DeleteHashTable(classes);
}

/* ---------------------------------------------------------------------- */

/* Install everything... */

int
Dkfobj_Init(interp)
     Tcl_Interp *interp;
{
  static int installed_types = 0;
  Tcl_HashTable *classes;

  if (!installed_types) {
    Tcl_RegisterObjType(&commandNameType);
    Tcl_RegisterObjType(&objectType);
    Tcl_RegisterObjType(&dereferenceMarkerType);
    installed_types = 1;
  }

  classes = (Tcl_HashTable *)Tcl_Alloc(sizeof(Tcl_HashTable));
  Tcl_InitHashTable(classes,TCL_STRING_KEYS);
  Tcl_CreateObjCommand(interp, "class", ClassCommand, classes, DelClassCmd);
  Tcl_SetVar(interp, "dkfobj_version", "2.0", TCL_GLOBAL_ONLY);
  Tcl_SetVar(interp, "dkfobj_patchLevel", "2.0a2", TCL_GLOBAL_ONLY);
  Tcl_SetVar(interp, "dkfobj_rcsinfo", rcsid, TCL_GLOBAL_ONLY);
  return TCL_OK;
}
